"""Methods to interact with SRA"""

import gzip
import os
import re
import subprocess
import sys
import warnings
from functools import partial
from multiprocessing import Pool
from subprocess import PIPE

import numpy as np
import pandas as pd
from tqdm.autonotebook import tqdm
from tqdm.contrib.concurrent import process_map
from tqdm.contrib.concurrent import thread_map

from .basedb import BASEdb
from .download import download_file
from .download import get_file_size
from .download import millify
from .filter_attrs import expand_sample_attribute_columns
from .taxid2name import TAXID_TO_NAME
from .utils import _find_aspera_keypath
from .utils import _get_url
from .utils import confirm
from .utils import copyfileobj
from .utils import get_gzip_uncompressed_size
from .utils import mkdir_p
from .utils import order_dataframe
from .utils import path_leaf
from .utils import run_command
from .utils import unique

warnings.simplefilter(action="ignore", category=FutureWarning)


tqdm.pandas()


FTP_PREFIX = {
    "fasp": "anonftp@ftp-trace.ncbi.nlm.nih.gov:",
    "ftp": "ftp://ftp-trace.ncbi.nlm.nih.gov",
}
SRADB_URL = [
    "https://s3.amazonaws.com/starbuck1/sradb/SRAmetadb.sqlite.gz",
    "https://gbnci-abcc.ncifcrf.gov/backup/SRAmetadb.sqlite.gz",
]

ASCP_CMD_PREFIX = "ascp -k1 -T -l 300m -P33001 -i"
PY3_VERSION = sys.version_info.minor


def _handle_download(record, use_ascp=False, pbar=None, ascp_bin=None, ascp_dir=None):
    srp = record["study_accession"]
    srx = record["experiment_accession"]
    srr = record["run_accession"]
    download_url = record["download_url"]
    srapath_url = record["srapath_url"]
    out_dir = record["out_dir"]
    if pbar:
        pbar.set_description("{}/{}/{}".format(srp, srx, srr))
    srp_dir = os.path.join(out_dir, srp)
    srx_dir = os.path.join(srp_dir, srx)
    download_filename = path_leaf(download_url)
    if ".fastq.gz" not in download_filename:
        srr_location = os.path.join(srx_dir, srr + ".sra")
    else:
        srr_location = os.path.join(srx_dir, download_filename)
    mkdir_p(srx_dir)
    if use_ascp:
        ascp = ASCP_CMD_PREFIX.replace("ascp", ascp_bin)
        ena_cols = [x for x in list(record.keys()) if "ena_fastq_ftp" in x]
        for col in ena_cols:
            download_url = record[col]
            cmd = "{} {} {} {}".format(
                ascp, _find_aspera_keypath(ascp_dir), download_url, srx_dir
            )
            run_command(cmd, verbose=False)
    else:
        if srapath_url is not None:
            download_filename = path_leaf(srapath_url)
            if ".fastq.gz" not in download_filename:
                srr_location = os.path.join(srx_dir, srr + ".sra")
            else:
                srr_location = os.path.join(srx_dir, download_filename)
            download_file(srapath_url, srr_location)
        else:
            download_file(download_url, srr_location)
    if pbar:
        pbar.update()


def _create_query(select_type_sql, gses):
    sql = (
        "SELECT DISTINCT "
        + select_type_sql
        + " FROM sra_ft WHERE sra_ft MATCH '"
        + " OR ".join(gses)
        + "';"
    )
    return sql


def _expand_sample_attrs(metadata_df):
    if "sample_attribute" in metadata_df.columns.tolist():
        metadata_df = expand_sample_attribute_columns(metadata_df)
        metadata_df = metadata_df.drop(columns=["sample_attribute"])
    return metadata_df


def _listify(ids):
    """convert string to list of unit length"""
    if isinstance(ids, str):
        ids = [ids]
    return ids


def _prettify_df(df, out_type, expand_sample_attributes):
    if len(df.index):
        df = df[out_type].sort_values(by=out_type)
    if expand_sample_attributes:
        df = _expand_sample_attrs(df)
    return df


def download_sradb_file(download_dir=os.getcwd(), overwrite=True, keep_gz=False):
    """Download SRAdb.sqlite file.

    Parameters
    ----------
    download_dir: string
                  Directory to download SRAmetadb.sqlite
    overwrite: bool
               overwrite existing file(s).
               Set to True by default.
    keep_gz: bool
             Delete .gz file after extraction is complete

    """
    download_location = os.path.join(download_dir, "SRAmetadb.sqlite.gz")
    download_location_unzip = download_location.rstrip(".gz")

    if os.path.isfile(download_location) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".format(
                download_location
            )
        )
    if os.path.isfile(download_location_unzip) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".format(
                download_location_unzip
            )
        )
    if os.path.isfile(download_location_unzip):
        os.remove(download_location_unzip)
    if os.path.isfile(download_location):
        os.remove(download_location)
    try:
        _get_url(SRADB_URL[0], download_location)
    except Exception as e:
        # Try other URL
        sys.stderr.write(
            "Could not use {}.\nException: {}.\nTrying alternate url ...\n".format(
                SRADB_URL[0], e
            )
        )
        _get_url(SRADB_URL[1], download_location)
    print("Extracting {} ...".format(download_location))
    filesize = get_gzip_uncompressed_size(download_location)
    with gzip.open(download_location, "rb") as fh_in:
        with open(download_location_unzip, "wb") as fh_out:
            copyfileobj(
                fh_in,
                fh_out,
                filesize=filesize,
                desc="Extracting {}".format("SRAmetadb.sqlite.gz"),
            )
    if not keep_gz:
        if os.path.isfile(download_location):
            os.remove(download_location)
    print("Done!")
    db = SRAdb(download_location_unzip)
    metadata = db.query("SELECT * FROM metaInfo")
    db.close()
    print("Metadata associated with {}:".format(download_location_unzip))
    print(metadata)


def _verify_srametadb(filepath):
    """Check if supplied SQLite is valid"""
    try:
        db = BASEdb(filepath)
    except:
        print(
            "{} not a valid SRAmetadb.sqlite file or path.\n".format(filepath)
            + "Please download one using `pysradb metadb`."
        )
        sys.exit(1)
    metadata = db.query("SELECT * FROM metaInfo")
    db.close()
    if list(metadata.iloc[0].values) != ["schema version", "1.0"]:
        print(
            "{} not a valid SRAmetadb.sqlite file.\n".format(filepath)
            + "Please download one using `pysradb metadb`."
        )
        sys.exit(1)


class SRAdb(BASEdb):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        _verify_srametadb(sqlite_file)
        super(SRAdb, self).__init__(sqlite_file)
        self._db_type = "SRA"
        self.valid_in_acc_type = [
            "SRA",
            "ERA",
            "DRA",
            "SRP",
            "ERP",
            "DRP",
            "SRS",
            "ERS",
            "DRS",
            "SRX",
            "ERX",
            "DRX",
            "SRR",
            "ERR",
            "DRR",
        ]
        self.valid_in_type = {
            "SRA": "submission",
            "ERA": "submission",
            "DRA": "submission",
            "SRP": "study",
            "ERP": "study",
            "DRP": "study",
            "SRS": "sample",
            "ERS": "sample",
            "DRS": "sample",
            "SRX": "experiment",
            "ERX": "experiment",
            "DRX": "experiment",
            "SRR": "run",
            "ERR": "run",
            "DRR": "run",
        }

    def sra_metadata(
        self,
        acc,
        out_type=[
            "study_accession",
            "experiment_accession",
            "sample_accession",
            "run_accession",
        ],
        assay=False,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
        output_read_lengths=False,
        acc_is_searchstr=False,
    ):
        """Get metadata for the provided SRA accession.

        Parameters
        ----------
        acc: string or list
             SRA accession ID
        out_type: list
                  List of columns to output
        assay: bool
               True if assay should be outputted
        sample_attribute: bool
                          True if sample_attribute should be outputted
        detailed: bool
                  True if full metadata tables should be outputted
        expand_sample_attributes: bool
                                  Should sample_attribute column be expanded?
        output_read_lengths: bool
                             True if read lengths should be calculated
        acc_is_searchstr: bool
                          True if acc is a search string


        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with all relevant columns
        """
        if not isinstance(acc, list):
            acc = [acc]
        for single_acc in acc:
            in_acc_type = re.sub("\\d+$", "", single_acc).upper()
            if in_acc_type not in self.valid_in_acc_type and not acc_is_searchstr:
                raise ValueError("{} not a valid input type".format(in_acc_type))
            if acc_is_searchstr:
                in_type = "study"
            else:
                in_type = self.valid_in_type[in_acc_type]
        output_columns = out_type[:]
        if detailed:
            output_columns += [
                "experiment_title",
                "experiment_attribute",
                "sample_attribute",
                "run_accession",
                "taxon_id",
                "library_selection",
                "library_layout",
                "library_strategy",
                "library_source",
                "library_name",
                "bases",
                "spots",
                "adapter_spec",
            ]
        if assay:
            if "library_strategy" not in output_columns:
                output_columns += ["library_strategy"]
        if sample_attribute:
            if "sample_attribute" not in output_columns:
                output_columns += ["sample_attribute"]
        output_columns = [x for x in output_columns if x != in_type]
        output_columns = unique(output_columns)
        select_type = [in_type + "_accession"] + output_columns
        select_type_sql = (",").join(select_type)
        sql = (
            "SELECT DISTINCT "
            + select_type_sql
            + " FROM sra_ft WHERE sra_ft MATCH '"
            + " OR ".join(acc)
            + "';"
        )
        df = self.query(sql)
        if not len(df.index):
            sys.stderr.write("Empty results")
            return df
        if "bases" in df.columns:
            if "spots" in df.columns:
                df["avg_read_length"] = df["bases"] / df["spots"]
                df["spots"] = df["spots"].astype(int)
            df["bases"] = df["bases"].astype(int)
        if "taxon_id" in df.columns:
            df["taxon_id"] = df["taxon_id"].fillna(0).astype(int)
            df = df.sort_values(by=["taxon_id"])
            df["organism_name"] = df["taxon_id"].apply(
                lambda taxid: TAXID_TO_NAME[taxid]
            )
            output_columns += ["organism_name"]
        if "experiment_accession" in df.columns and "run_accession" in df.columns:
            df = df.sort_values(by=["experiment_accession", "run_accession"])
        elif "experiment_accession" in df.columns:
            df = df.sort_values(by=["experiment_accession"])
        elif "run_accession" in df.columns:
            df = df.sort_values(by=["run_accession"])
        elif "sample_accession" in df.columns:
            df = df.sort_values(by=["sample_accession"])
        if output_read_lengths and "avg_read_length" in df.columns:
            output_columns = output_columns + ["avg_read_length"]
        metadata_df = df.reset_index(drop=True)
        metadata_df = order_dataframe(metadata_df, output_columns)
        if expand_sample_attributes:
            if "sample_attribute" in metadata_df.columns.tolist():
                metadata_df = expand_sample_attribute_columns(metadata_df)
                metadata_df = metadata_df.drop(columns=["sample_attribute"])
        return metadata_df

    def srp_to_srx(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRP to SRX/SRR.

        Parameters
        ----------
        srp: string
             SRP ID

        Returns
        -------
        srp_to_srx_df: DataFrame
                       DataFrame with two columns for SRX/SRR
        """
        out_type = ["experiment_accession"]
        if detailed:
            out_type += [
                "sample_accession",
                "run_accession",
                "experiment_alias",
                "sample_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        return self.sra_metadata(
            acc=srp,
            out_type=out_type,
            sample_attribute=sample_attribute,
            expand_sample_attributes=expand_sample_attributes,
        )

    def srp_to_srs(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRP to SRS.

        Parameters
        ----------
        srp: string
             SRP ID

        Returns
        -------
        srp_to_srs_df: DataFrame
                       DataFrame with two columns for SRS
        """
        out_type = ["study_accession", "sample_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "run_accession",
                "study_alias",
                "sample_alias",
                "experiment_alias",
                "run_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        df = self.sra_metadata(
            acc=srp,
            out_type=out_type,
            expand_sample_attributes=expand_sample_attributes,
        )
        return df

    def srp_to_srr(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRP to SRR.

        Parameters
        ----------
        srp: string
             SRP ID

        Returns
        -------
        srp_to_srr_df: DataFrame
        """
        out_type = ["study_accession", "run_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "study_alias",
                "experiment_alias",
                "sample_alias",
                "run_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        df = self.sra_metadata(
            acc=srp,
            out_type=out_type,
            expand_sample_attributes=expand_sample_attributes,
        )
        return df

    def srp_to_gse(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRP to GSE

        Parameters
        ----------
        srp: string
             SRP ID

        Returns
        -------
        srp_to_srr_df: DataFrame
        """
        out_type = ["study_accession", "study_alias"]
        if detailed:
            out_type += [
                "experiment_accession",
                "run_accession",
                "sample_accession",
                "experiment_alias",
                "run_alias",
                "sample_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        return self.sra_metadata(
            acc=srp,
            out_type=out_type,
            expand_sample_attributes=expand_sample_attributes,
        )

    def gse_to_srp(
        self,
        gses,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRP to GSE

        Parameters
        ----------
        gses: string or list
              List of GSE ID

        Returns
        -------
        gse_to_srp_df: DataFrame
        """
        gses = _listify(gses)
        out_type = ["study_alias", "study_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "experiment_alias",
                "sample_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gses)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gsm_to_srp(
        self,
        gsms,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSM to SRP.

        Parameters
        ----------
        gsms: string or list
              List of GSM ID

        Returns
        -------
        gsm_to_srp_df: DataFrame
        """
        gsms = _listify(gsms)
        out_type = ["experiment_alias", "study_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "run_accession",
                "experiment_alias",
                "sample_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gsms)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gsm_to_srr(
        self,
        gsms,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSMs to SRR.

        Parameters
        ----------
        gsms: string or list
              List of GSM id
        sample_attribute: bool
                          Include `sample_attribute` column


        Returns
        -------
        gsm_to_srr_df: DataFrame
                       DataFrame with two columns for GSM/SRR
        """
        gsms = _listify(gsms)

        out_type = ["experiment_alias", "run_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "study_accession",
                "run_alias",
                "sample_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gsms)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gsm_to_srs(
        self,
        gsms,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSM to SRS.

        Parameters
        ----------
        gsms: list or string
              List of gsms

        Returns
        -------
        gsm_to_srs_df: DataFrame
        """
        gsms = _listify(gsms)
        out_type = ["sample_alias", "sample_accession"]
        if detailed:
            out_type += [
                "sample_accession",
                "experiment_accession",
                "run_accession",
                "study_accession",
                "sample_alias",
                "experiment_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gsms)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gsm_to_srx(
        self,
        gsms,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSM to SRX.

        Parameters
        ----------
        srx: string
             SRX ID

        Returns
        -------
        srs_to_srx_df: DataFrame
        """
        gsms = _listify(gsms)
        out_type = ["experiment_alias", "experiment_accession"]
        if detailed:
            out_type += [
                "sample_accession",
                "run_accession",
                "study_accession",
                "sample_alias",
                "experiment_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gsms)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gse_to_gsm(
        self,
        gses,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSE to GSM

        Parameters
        ----------
        gses: string or list
              List of GSE ID

        Returns
        -------
        gse_to_gsm_df: DataFrame
        """
        gses = _listify(gses)
        out_type = ["study_alias", "experiment_alias"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "run_accession",
                "sample_alias",
                "run_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gses)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def gsm_to_gse(
        self,
        gsms,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert GSM to GSE

        Parameters
        ----------
        gsms: string or list
              List of GSM ID

        Returns
        -------
        gsm_to_gse_df: DataFrame
        """
        gsms = _listify(gsms)
        out_type = ["experiment_alias", "study_alias"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "run_accession",
                "sample_alias",
                "run_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, gsms)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srr_to_srp(
        self,
        srrs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRR to SRP.

        Parameters
        ----------
        srr: list of string
             List of SRR IDs

        Returns
        -------
        srr_to_srp_df: DataFrame
        """
        srrs = _listify(srrs)
        out_type = ["run_accession", "study_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "sample_accession",
                "run_alias",
                "study_alias",
                "experiment_alias",
                "sample_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srrs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srr_to_srs(
        self,
        srrs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRR to SRS.

        Parameters
        ----------
        srr: list of string
             List of SRR IDs

        Returns
        -------
        srp_to_srs_df: DataFrame
        """
        srrs = _listify(srrs)
        out_type = ["run_accession", "sample_accession"]
        if detailed:
            out_type += [
                "experiment_accession",
                "study_accession",
                "run_alias",
                "sample_alias",
                "experiment_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srrs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srx_to_srs(
        self,
        srxs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRX to SRS.

        Parameters
        ----------
        srx: string
             SRX ID

        Returns
        -------
        srp_to_srs_df: DataFrame
        """
        srxs = _listify(srxs)
        out_type = ["experiment_accession", "sample_accession"]
        if detailed:
            out_type += [
                "run_accession",
                "study_accession",
                "experiment_alias",
                "sample_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srxs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srs_to_gsm(
        self,
        srss,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRS to GSM.

        Parameters
        ----------
        srss: list or string
              List of SRS ID

        Returns
        -------
        srs_to_gsm_df: DataFrame
        """
        srss = _listify(srss)
        out_type = ["sample_accession", "sample_alias"]
        if detailed:
            out_type += [
                "experiment_accession",
                "run_accession",
                "study_accession",
                "experiment_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srss)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srs_to_srx(
        self,
        srss,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRS to SRX.

        Parameters
        ----------
        srx: string
             SRX ID

        Returns
        -------
        srs_to_srx_df: DataFrame
        """
        srss = _listify(srss)
        out_type = ["sample_accession", "experiment_accession"]
        if detailed:
            out_type += [
                "run_accession",
                "study_accession",
                "sample_alias",
                "experiment_alias",
                "run_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srss)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srr_to_srx(
        self,
        srrs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRR to SRX.

        Parameters
        ----------
        srrs: string or list
              List of SRR id
        sample_attribute: bool
                          Include `sample_attribute` column


        Returns
        -------
        srr_to_srx_df: DataFrame
                       DataFrame with two columns for SRX/SRR
        """
        srrs = _listify(srrs)

        out_type = ["run_accession", "experiment_accession"]
        if detailed:
            out_type += [
                "sample_accession",
                "study_accession",
                "run_alias",
                "experiment_alias",
                "sample_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srrs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srx_to_srp(
        self,
        srxs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRXs to SRP.

        Parameters
        ----------
        srxs: string or list
              List of SRX id
        sample_attribute: bool
                          Include `sample_attribute` column


        Returns
        -------
        srx_to_srp_df: DataFrame
                       DataFrame with two columns for SRX
        """
        srxs = _listify(srxs)
        out_type = ["experiment_accession", "study_accession"]
        if detailed:
            out_type += [
                "run_accession",
                "sample_accession",
                "experiment_alias",
                "run_alias",
                "sample_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srxs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srr_to_gsm(
        self,
        srrs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRR to GSM

        Parameters
        ----------
        gses: string or list
              List of SRR

        Returns
        -------
        srr_to_gsm_df: DataFrame
        """
        srrs = _listify(srrs)
        out_type = ["run_alias", "experiment_alias"]
        if detailed:
            out_type += [
                "run_accession",
                "experiment_accession",
                "study_accession",
                "sample_accession",
                "study_alias" "sample_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srrs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def srx_to_srr(
        self,
        srxs,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
    ):
        """Convert SRXs to SRR/SRP.

        Parameters
        ----------
        srxs: string or list
              List of SRX id
        sample_attribute: bool
                          Include `sample_attribute` column


        Returns
        -------
        srx_to_srp_df: DataFrame
                       DataFrame with two columns for SRX/SRR
        """
        srxs = _listify(srxs)
        out_type = ["experiment_accession", "run_accession"]
        if detailed:
            out_type += [
                "sample_accession",
                "study_accession",
                "experiment_alias",
                "run_alias",
                "sample_alias",
                "study_alias",
            ]
        if sample_attribute:
            out_type += ["sample_attribute"]
        select_type_sql = (",").join(out_type)
        sql = _create_query(select_type_sql, srxs)
        df = self.query(sql)
        df = _prettify_df(df, out_type, expand_sample_attributes)
        return df

    def search_sra(
        self,
        search_str,
        out_type=[
            "study_accession",
            "experiment_accession",
            "sample_accession",
            "run_accession",
        ],
        assay=False,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
        output_read_lengths=False,
    ):
        """Search SRA for any search term.

        Parameters
        ----------
        search_str: string
                    SQL like text string to search.
                    SQL like text => For example,
                    terms in quotes "" enforce an exact search.


        Returns
        -------
        query_df: DataFrame
                  Dataframe with relevant query results
        """
        return self.sra_metadata(
            search_str,
            out_type=out_type,
            assay=assay,
            sample_attribute=sample_attribute,
            detailed=detailed,
            expand_sample_attributes=expand_sample_attributes,
            output_read_lengths=output_read_lengths,
            acc_is_searchstr=True,
        )

    def search_by_expt_id(self, srx):
        """Search for a SRX/GSM id in the experiments.

        Parameters
        ----------
        srx: string
             SRX (experiment_accession) ID

        Returns
        -------
        results: dict
                 Dictionary with relevant hits
        """
        if "GSM" in srx:
            results = self.cursor.execute(
                'select * from EXPERIMENT where experiment_alias = "{}"'.format(srx)
            ).fetchall()
        else:
            results = self.cursor.execute(
                'select * from EXPERIMENT where experiment_accession = "{}"'.format(srx)
            ).fetchall()
        assert len(results) == 1, "Got multiple hits"
        results = results[0]
        column_names = list([x[0] for x in self.cursor.description])
        results = dict(list(zip(column_names, results)))
        return pd.DataFrame.from_dict(results, orient="index").T

    @staticmethod
    def _srapath_url(sacc):
        """Get srapath URL for a SRP/SRR.

        Parameters
        ----------
        srp: string


        Returns
        -------
        srr_paths: dict
                   A dict of URLS with keys as SRR
        """
        if PY3_VERSION >= 7:
            proc = subprocess.run(["srapath", sacc], capture_output=True)
        else:
            proc = subprocess.run(["srapath", sacc], stdout=PIPE, stderr=PIPE)
        stdout = str(proc.stdout.strip().decode("utf-8"))

        urls = stdout.split("\n")
        # TODO: Improve this
        if not urls[0]:
            sys.stderr.write("Unable to run srapath.\n")
            return None
        urls = list(map(str, urls))
        srrs = [str(url.strip().split("/")[-1].split(".")[0]) for url in urls]
        return dict(zip(srrs, urls))

    @staticmethod
    def _srapath_url_srr(srr):
        """Get srapath URL for a SRR.

        Parameters
        ----------
        srp: string


        Returns
        -------
        srr_paths: dict
                   A dict of URLS with keys as SRR
        """
        if PY3_VERSION >= 7:
            proc = subprocess.run(["srapath", srr], capture_output=True)
        else:
            proc = subprocess.run(["srapath", srr], stdout=PIPE, stderr=PIPE)
        stdout = str(proc.stdout.strip().decode("utf-8"))
        urls = stdout.split("\n")
        # TODO: Improve this
        if not urls[0]:
            sys.stderr.write("Unable to run srapath.\n")
            return None
        return str(urls[0])

    def _select_best_url(self, url_list, row, use_ascp):
        if use_ascp:
            for url in url_list:
                if str(row[url]).startswith("fasp"):
                    return row[url]
        else:
            for url in url_list:
                if str(row[url]).startswith("http"):
                    return row[url]
            for url in url_list:
                if str(row[url]).startswith("ftp"):
                    return row[url]
        return None

    def _format_dataframe_for_download(self, df, url_column, use_ascp):
        """Format a dataframe as input for pysradb download.

        This method formats the input dataframe into the sradb.download
        method.
        First, the columns "study_accession", "experiment_accession", and
        "run_accession" will be identified.
        Next, this method will attempt to find the download url
        corresponding to each run_accession.
        If url_column is supplied, the column will be used if found.
        If url_column is not supplied or not found, the method looks
        for column headers in the dataframe that matches the regex
        string ".*sra.*(url|ftp|galaxy).*", which looks for a column header
        that contains "sra", as well as either "url", "ftp" or "galaxy"
        after "sra". (case insensitive). If use_ascp is true, the program
        will look for known columns containing ascp links instead.

        All matching columns will be
        returned for the user to select the url column to use. Additionally,
        the column "recommended_url" is returned as well, which ensures as
        much as possible that the column contains url links for every run
        accession.
        If any of "study_accession", "experiment_accession",
        "run_accession" or "srapath_url" is missing, a
        MissingDataFrameColumnsException will be raised.

        Parameters
        ----------
        df: Pandas.DataFrame
            dataframe containing accession numbers of interest as well as
            potentially other metadata pertaining to the accession numbers.
        url_column: str
            name of the dataframe column header that contains download
            urls, or a regex matching the expected column header.
            if url_column == None, the regex ".*sra.*(url|ftp|galaxy).*"
            will be used.
        use_ascp: bool
            whether ascp is used. If true and url_column is invalid or not
            specified, the program will look for known columns containing
            ascp links (containing "aspera" or "fasta")

        Returns
        -------
        df_for_download: Pandas.DataFrame
            dataframe containing accession numbers and (preferably) URL links
            that can be used for download.
        """
        missing_columns = []
        df_columns = df.columns.tolist()
        accession_columns = ["study_accession", "experiment_accession", "run_accession"]
        for accession_column in accession_columns:
            if accession_column not in df_columns:
                missing_columns.append(accession_column)
        # Special case for SraSearch
        run_count = 1  # each row in the df contains at most 1 run_accession
        while f"run_{run_count}_accession" in df_columns:
            run_count += 1
        if missing_columns == ["run_accession"] and run_count > 1:
            missing_columns.clear()
            accession_columns[-1] = "run_1_accession"
        if url_column and url_column in df_columns:
            formatted_df = df.loc[
                :, df.columns.isin(accession_columns + [url_column])
            ].copy()
            formatted_df.rename(
                {"run_1_accession": "run_accession"},
                inplace=True,
            )
        elif use_ascp and run_count == 1:
            # Add aspera columns, if they exist(for EnaSearch/metadata)
            possible_aspera_cols = [
                "fastq_aspera",
                "sra_aspera",
                "submitted_aspera",
                "ena_fastq_ftp",
            ]
            aspera_cols = []
            for col in possible_aspera_cols:
                if col in df_columns:
                    aspera_cols.append(col)
            formatted_df = (
                df.loc[:, df.columns.isin(accession_columns + aspera_cols)]
                .rename({"run_1_accession": "run_accession"})
                .copy()
            )

        else:
            run_dfs = []
            url_regex = re.compile(".*sra.*(url|ftp|galaxy).*", re.IGNORECASE)
            matched_cols = list(filter(url_regex.match, df_columns))
            for i in range(1, run_count):
                url_list = list(
                    filter(lambda x: x.startswith(f"run_{i}"), matched_cols)
                )
                df["recommended_url"] = df.apply(
                    lambda row: self._select_best_url(url_list, row, use_ascp), axis=1
                )

                def remove_unusable_urls(url, use_ascp):
                    if use_ascp:
                        if url.startswith("fasp"):
                            return url
                    else:
                        if url.startswith("http") or url.startswith("ftp"):
                            return url
                    return None

                for url_c in url_list:
                    df[url_c] = df[url_c].apply(
                        lambda url: remove_unusable_urls(url, use_ascp)
                    )
                expected_columns = [
                    "study_accession",
                    "experiment_accession",
                    f"run_{i}_accession",
                    "recommended_url",
                ] + url_list
                run_df = df.loc[:, df.columns.isin(expected_columns)].copy()
                run_df = run_df.rename(columns={f"run_{i}_accession": "run_accession"})
                run_dfs.append(run_df)
            if run_count == 1:
                expected_columns = [
                    "study_accession",
                    "experiment_accession",
                    "run_accession",
                    "recommended_url",
                ]
                df["recommended_url"] = df.apply(
                    lambda row: self._select_best_url(matched_cols, row, use_ascp),
                    axis=1,
                ).tolist()

                run_dfs = [df.loc[:, df.columns.isin(expected_columns)].copy()]
            formatted_df = pd.concat(run_dfs)
            formatted_df.dropna(axis=1, how="all")
            if not matched_cols:
                print(
                    f"No URL column is found.\n"
                    "You may wish to re-run your query with either\n"
                    "pysradb metadata --detailed \n"
                    "or \n"
                    "pysradb search -v 3\n"
                    "Generating default download URL for each run accession...\n",
                    flush=True,
                )
        if missing_columns:
            sys.stderr.write(
                "\npysradb download is unable to run:\n"
                "The following required columns are missing from the input DataFrame:\n"
                f"{missing_columns}\n\n"
                "Please run your query with either\n"
                "pysradb metadata --detailed \n"
                "or \n"
                "pysradb search --detailed\n"
                "or \n"
                "pysradb search -v 3\n"
            )
            sys.exit(1)
        return formatted_df.dropna(
            subset=["study_accession", "experiment_accession", "run_accession"]
        )

    def download(
        self,
        srp=None,
        df=None,
        url_col="sra_url",
        out_dir=None,
        filter_by_srx=[],
        use_ascp=False,
        ascp_dir=None,
        ascp_bin=None,
        skip_confirmation=False,
        threads=1,
    ):
        """Download SRA files.

        Parameters
        ----------
        srp: string
             SRP ID (optional)
        df: Dataframe
            A dataframe as obtained from `sra_metadata`
        url_col: string
                Column of df to use for downloading
        out_dir: string
                 Directory location for download
        filter_by_srx: list
                       List of SRX ids to filter
        protocol: string
                  ['fasp'/'ftp'] fasp => faster download, ftp => slower
        ascp_dir: string
                  Location of ascp directory
        """
        if out_dir is None:
            out_dir = os.path.join(os.getcwd(), "pysradb_downloads")
        if srp:
            df = self.sra_metadata(srp, detailed=True)
        if use_ascp:
            if ascp_dir is None:
                ascp_dir = os.path.join(os.path.expanduser("~"), ".aspera")
            if not os.path.exists(ascp_dir):
                sys.stderr.write(
                    "Count not find aspera at: {}\n".format(ascp_dir)
                    + "Install aspera-client following instructions"
                    + "at https://github.com/saketkc/pysradb/README.rst for faster downloads.\n"
                    + "You can supress this message by using `--use-wget` flag\n"
                    + "Continuing with wget ...\n\n"
                )
                use_ascp = False
            else:
                ascp_bin = os.path.join(ascp_dir, "connect", "bin", "ascp")
        # Does the necessary column formatting for the dataframe
        df = self._format_dataframe_for_download(df.copy(), url_col, use_ascp)
        if url_col not in df.columns:
            print(f'The supplied url column "{url_col}" cannot be found.\n')
            url_col = "recommended_url"
            if not skip_confirmation:
                pd.set_option("display.max_colwidth", -1)
                print(df.to_string(index=False, justify="left", col_space=0))
                print(os.linesep, flush=True)
                if not confirm("Use recommended_url instead?"):
                    url_col = input("Please enter an url column to use:  ")
            else:
                print("Using recommended_url instead.\n", flush=True)
        if url_col not in df.columns:
            sys.exit("\nMissing url columns!")

        if filter_by_srx:
            if isinstance(filter_by_srx, str):
                filter_by_srx = [filter_by_srx]
        if filter_by_srx:
            df = df.loc[df.experiment_accession.isin(filter_by_srx)]
        df["download_url"] = (
            FTP_PREFIX["ftp"]
            + "/sra/sra-instant/reads/ByRun/sra/"
            + df["run_accession"].str[:3]
            + "/"
            + df["run_accession"].str[:6]
            + "/"
            + df["run_accession"]
            + "/"
            + df["run_accession"]
            + ".sra"
        ).tolist()
        ena_columns = [col for col in df.columns if "ena" in col]
        df["out_dir"] = out_dir
        if not len(df.index):
            print("Could not locate {} in db".format(srp))
            sys.exit(0)
        if not use_ascp:
            print("Checking download URLs", flush=True)
            df["filesize"] = df.apply(
                lambda x: get_file_size(x, url_col), axis=1
            ).tolist()
            df.dropna(subset=["filesize"])
            total_file_size = millify(np.sum(df["filesize"]))
            df["filesize"] = df["filesize"].apply(lambda x: millify(x)).tolist()
            print("The following files will be downloaded: \n")
            pd.set_option("display.max_colwidth", -1)
            print(df.to_string(index=False, justify="left", col_space=0))
            print(os.linesep)
            print("Total size: {}".format(total_file_size))
            print(os.linesep, flush=True)
            if not skip_confirmation:
                if not confirm("Start download? "):
                    sys.exit(0)
        df["srapath_url"] = df[url_col].tolist()
        thread_map(
            partial(
                _handle_download,
                use_ascp=use_ascp,
                ascp_bin=ascp_bin,
                ascp_dir=ascp_dir,
            ),
            df.to_dict("records"),
            max_workers=threads,
        )

        return df



"""Tests for sradb.py
"""

import os
from sqlite3 import OperationalError

import pytest

from pysradb import SRAdb
from pysradb.filter_attrs import guess_cell_type
from pysradb.filter_attrs import guess_strain_type
from pysradb.filter_attrs import guess_tissue_type


@pytest.fixture(scope="module")
def sradb_connection(conf_download_sradb_file):
    db_file = conf_download_sradb_file
    db = SRAdb(db_file)
    return db


def test_list_tables(sradb_connection):
    sra_tables = sradb_connection.list_tables()
    assert sra_tables == [
        "metaInfo",
        "submission",
        "study",
        "sample",
        "experiment",
        "run",
        "sra",
        "sra_ft",
        "sra_ft_content",
        "sra_ft_segments",
        "sra_ft_segdir",
        "col_desc",
        "fastq",
    ]


def test_list_fields(sradb_connection):
    fields = sradb_connection.list_fields("study")
    assert fields == [
        "study_ID",
        "study_alias",
        "study_accession",
        "study_title",
        "study_type",
        "study_abstract",
        "broker_name",
        "center_name",
        "center_project_name",
        "study_description",
        "related_studies",
        "primary_study",
        "sra_link",
        "study_url_link",
        "xref_link",
        "study_entrez_link",
        "ddbj_link",
        "ena_link",
        "study_attribute",
        "submission_accession",
        "sradb_updated",
    ]


def test_desc_table(sradb_connection):
    names = sorted(sradb_connection.desc_table("sra_ft").name.tolist())
    assert names[:7] == [
        "SRR_bamFile",
        "SRX_bamFile",
        "SRX_fastqFTP",
        "adapter_spec",
        "anonymized_name",
        "base_caller",
        "bases",
    ]


def test_all_row_counts(sradb_connection):
    assert sradb_connection.all_row_counts().loc["metaInfo", "count"] == 2


def test_all_row_counts2(sradb_connection):
    assert len(sradb_connection.all_row_counts()) == 13


def test_sra_metadata(sradb_connection):
    df = sradb_connection.sra_metadata("SRP017942")
    assert df["experiment_accession"][0] == "SRX217027"


def test_sra_metadata2(sradb_connection):
    df = sradb_connection.sra_metadata(
        "SRP017942", detailed=True, expand_sample_attributes=True
    )
    assert "3xflag-gfp" in df["transfected_with"].tolist()


def test_search(sradb_connection):
    df = sradb_connection.search_sra(search_str="breast cancer")
    assert len(df.index)


def test_search2(sradb_connection):
    df = sradb_connection.search_sra(
        '"salivary microbiome" AND "diabetes mellitus"', detailed=True
    )
    assert "SRP241848" in df["study_accession"].to_list()


def test_search_by_expt_id(sradb_connection):
    df = sradb_connection.search_by_expt_id("SRX1254413")
    assert df.study_name.tolist()[0] == "GSE73136"


def test_search_by_expt_id2(sradb_connection):
    srx_id = "SRX116363"
    df_expt = sradb_connection.search_by_expt_id(srx_id)
    sra_id = df_expt["submission_accession"].loc[0]
    df = sradb_connection.sra_metadata(sra_id)
    connected_srp = sradb_connection.srx_to_srp("SRX116363").iloc[0, 1]
    assert (srx_id in df["experiment_accession"].to_list()) and (
        connected_srp == "SRP010374"
    )


# def test_download_fasp(sradb_connection):
#    df = sradb_connection.sra_metadata("SRP098789")
#    df = df[df.experiment_accession == "SRX2536403"]
#    sradb_connection.download(df=df, out_dir="data/", skip_confirmation=True)
#    assert os.path.isfile("data/SRP098789/SRX2536403/SRR5227288.sra")
#    assert os.path.getsize("data/SRP098789/SRX2536403/SRR5227288.sra")
#    os.remove("data/SRP098789/SRX2536403/SRR5227288.sra")


@pytest.mark.xfail
def test_download_ftp(sradb_connection):
    # This happens to fail because of ftp problems
    df = sradb_connection.sra_metadata("SRP098789")
    df = df[df.experiment_accession == "SRX2536404"]
    sradb_connection.download(
        df=df, protocol="ftp", out_dir="data/", skip_confirmation=True
    )
    assert os.path.isfile("data/SRP098789/SRX2536404/SRR5227289.sra")
    assert os.path.getsize("data/SRP098789/SRX2536404/SRR5227289.sra")
    os.remove("data/SRP098789/SRX2536404/SRR5227289.sra")


def test_tissue_type(sradb_connection):
    df = sradb_connection.sra_metadata("SRP016501", detailed=True)
    df = df[df.experiment_accession == "SRX196389"]
    cell_type = df["sample_attribute"].apply(lambda x: guess_cell_type(x))
    tissue_type = df["sample_attribute"].apply(lambda x: guess_tissue_type(x))
    assert cell_type.tolist() == ["chicken_brain"]
    assert tissue_type.tolist() == ["brain"]


def test_strain_type(sradb_connection):
    df = sradb_connection.sra_metadata("SRP043036", detailed=True)
    df = df.sort_values(by="experiment_accession")
    strains = df["sample_attribute"].apply(lambda x: guess_strain_type(x)).tolist()
    assert strains == [
        "by4741",
        "by4741",
        "by4741",
        "by4741",
        "by4741",
        "by4741",
        "by4741",
        "by4741",
        "s288c",
        "s288c",
        "s288c",
        "s288c",
    ]


def test_srp_to_srx(sradb_connection):
    assert len(sradb_connection.srp_to_srx("SRP082570")) == 14


def test_srp_to_srr(sradb_connection):
    df = sradb_connection.srp_to_srr("SRP091987")
    assert sorted(list(df["run_accession"])[:3]) == [
        "SRR4447104",
        "SRR4447105",
        "SRR4447106",
    ]


def test_srp_to_gse(sradb_connection):
    gse_id = sradb_connection.srp_to_gse("SRP050443").iloc[0, 1]
    df = sradb_connection.gse_to_gsm(gse_id)
    assert "GSM1557451" in df["experiment_alias"].to_list()


def test_gsm_to_gse(sradb_connection):
    df = sradb_connection.gsm_to_gse(["GSM1020651", "GSM1020664", "GSM1020771"])
    assert set(list(df["study_alias"])) == {"GSE41637"}


def test_srs_to_gsm(sradb_connection):
    df = sradb_connection.srs_to_gsm("SRS1757470")
    assert "GSM2358940" == df.iloc[0, 1]


@pytest.mark.xfail(raises=ValueError)
def test_wrong_input_metadata(sradb_connection):
    df = sradb_connection.sra_metadata("should_throw_error")



"""Tests for sradb.py
"""

import os
from sqlite3 import OperationalError

import pytest

from pysradb import SRAdb
from pysradb.filter_attrs import guess_cell_type
from pysradb.filter_attrs import guess_strain_type
from pysradb.filter_attrs import guess_tissue_type


def test_not_valid_file():
    """Test to check for error if file is either not
    present or not a valid sqlite file"""
    path = "SRAmetadb.sqlite"
    try:
        db = SRAdb(path)
    except SystemExit:
        assert os.path.isfile(path) == False
    except OperationalError:
        assert True



# -*- coding: utf-8 -*-
"""Top-level package for pysradb."""

__author__ = """Saket Choudhary"""
__email__ = "saketkc@gmail.com"
__version__ = "2.0.0"

from .filter_attrs import expand_sample_attribute_columns
from .geodb import GEOdb
from .geodb import download_geodb_file
from .sradb import SRAdb
from .sradb import download_sradb_file
from .sraweb import SRAweb



"""Utilities to interact with SRA online"""

import concurrent.futures
import os
import sys
import time
import warnings
from collections import OrderedDict
from json.decoder import JSONDecodeError
from xml.parsers.expat import ExpatError

import numpy as np
import pandas as pd
import requests
import xmltodict

from .sradb import SRAdb

warnings.simplefilter(action="ignore", category=FutureWarning)

from xml.sax.saxutils import escape


def xmlescape(data):
    return escape(data, entities={"'": "&apos;", '"': "&quot;"})


def _order_first(df, column_order_list):
    columns = column_order_list + [
        col for col in df.columns.tolist() if col not in column_order_list
    ]
    # check if all columns do exist in the dataframe
    if len(set(columns).intersection(df.columns)) == len(columns):
        df = df.loc[:, columns]
    return df


def _retry_response(base_url, payload, key, max_retries=10):
    """Rerty fetching esummary if API rate limit exceeeds"""
    for index, _ in enumerate(range(max_retries)):
        try:
            request = requests.get(base_url, params=OrderedDict(payload))
            response = request.json()
            results = response[key]
            return response
        except KeyError:
            # sleep for increasing times
            time.sleep(index + 1)
            continue
    raise RuntimeError("Failed to fetch esummary. API rate limit exceeded.")


def get_retmax(n_records, retmax=500):
    """Get retstart and retmax till n_records are exhausted"""
    for i in range(0, n_records, retmax):
        yield i


class SRAweb(SRAdb):
    def __init__(self, api_key=None):
        """
        Initialize a SRAwebdb.

        Parameters
        ----------

        api_key: string
                 API key for ncbi eutils.
        """
        self.base_url = dict()
        self.base_url[
            "esummary"
        ] = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
        self.base_url[
            "esearch"
        ] = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
        self.base_url[
            "efetch"
        ] = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"

        self.ena_fastq_search_url = "https://www.ebi.ac.uk/ena/portal/api/filereport"
        self.ena_params = [("result", "read_run"), ("fields", "fastq_ftp")]

        self.esearch_params = {}
        self.esearch_params["sra"] = [
            ("db", "sra"),
            ("usehistory", "n"),
            ("retmode", "json"),
        ]
        self.esearch_params["geo"] = [
            ("db", "gds"),
            ("usehistory", "n"),
            ("retmode", "json"),
        ]
        self.efetch_params = [
            ("db", "sra"),
            ("usehistory", "n"),
            ("retmode", "runinfo"),
        ]

        if api_key is not None:
            self.esearch_params["sra"].append(("api_key", str(api_key)))
            self.esearch_params["geo"].append(("api_key", str(api_key)))
            self.efetch_params.append(("api_key", str(api_key)))
            self.sleep_time = 1 / 10
        else:
            self.sleep_time = 1 / 3

    @staticmethod
    def format_xml(string):
        """Create a fake root to make 'string' a valid xml

        Parameters
        ----------
        string: str

        Returns
        --------
        xml: str
        """
        # string = unescape(string.strip())
        string = string.strip()
        return "<root>" + string + "</root>"

    @staticmethod
    def xml_to_json(xml):
        """Convert xml to json.

        Parameters
        ----------
        xml: str
             Input XML

        Returns
        -------
        xml_dict: dict
                  Parsed xml as dict
        """
        try:
            xmldict = xmltodict.parse(
                xml, process_namespaces=True, dict_constructor=OrderedDict
            )
            json = xmldict["root"]
        except ExpatError:
            raise RuntimeError("Unable to parse xml: {}".format(xml))
        return json

    def fetch_ena_fastq(self, srp):
        """Fetch FASTQ records from ENA (EXPERIMENTAL)

        Parameters
        ----------
        srp: string
             Srudy accession

        Returns
        -------
        srr_url: list
                 List of SRR fastq urls
        """
        payload = self.ena_params.copy()
        payload += [("accession", srp)]
        request = requests.get(self.ena_fastq_search_url, params=OrderedDict(payload))
        request_text = request.text.strip()
        urls = []
        for line in request_text.split("\n"):
            if "fastq_ftp" in line:
                continue
            line = line.strip()
            line_split = line.split("\t")
            if len(line_split) != 2:
                continue
            srr, url = line.split("\t")
            http_url = "http://{}".format(url)
            ftp_url = url.replace("ftp.sra.ebi.ac.uk/", "era-fasp@fasp.sra.ebi.ac.uk:")
            urls += [(srr, http_url, ftp_url)]

        # Paired end case
        def _handle_url_split(url_split):
            url1_1 = pd.NA
            url1_2 = pd.NA
            for url_temp in url_split:
                if "_1.fastq.gz" in url_temp:
                    url1_1 = url_temp
                elif "_2.fastq.gz" in url_temp:
                    url1_2 = url_temp
            return url1_1, url1_2

        if ";" in request_text:
            urls_expanded = []
            for srr, url1, url2 in urls:
                # strip _1, _2
                srr = srr.split("_")[0]
                if ";" in url1:
                    url1_split = url1.split(";")
                    if len(url1_split) == 2:
                        url1_1, url1_2 = url1_split
                    else:
                        # warnings.warn('ignoring extra urls found for paired end accession')
                        url1_1, url1_2 = _handle_url_split(url1_split)
                    url1_2 = "http://{}".format(url1_2)
                    url2_split = url2.split(";")
                    if len(url2_split) == 2:
                        url2_1, url2_2 = url2_split
                    else:
                        # warnings.warn('ignoring extra urls found for paired end accession')
                        url2_1, url2_2 = _handle_url_split(url2_split)
                else:
                    url1_1 = url1
                    url2_1 = url2
                    url1_2 = ""
                    url2_2 = ""
                urls_expanded.append((srr, url1_1, url1_2, url2_1, url2_2))
            return pd.DataFrame(
                urls_expanded,
                columns=[
                    "run_accession",
                    "ena_fastq_http_1",
                    "ena_fastq_http_2",
                    "ena_fastq_ftp_1",
                    "ena_fastq_ftp_2",
                ],
            ).sort_values(by="run_accession")
        else:
            return pd.DataFrame(
                urls, columns=["run_accession", "ena_fastq_http", "ena_fastq_ftp"]
            ).sort_values(by="run_accession")

    def create_esummary_params(self, esearchresult, db="sra"):
        query_key = esearchresult["querykey"]
        webenv = esearchresult["webenv"]
        retstart = esearchresult["retstart"]

        # TODO this should be adaptive to build
        # upon using the 'count' result in esearch result,
        # Currently only supports a max of 500 records.
        # retmax = esearchresult["retmax"]
        retmax = 500

        return [
            ("query_key", query_key),
            ("WebEnv", webenv),
            ("retstart", retstart),
            ("retmax", retmax),
        ]

    def get_esummary_response(self, db, term, usehistory="y"):
        assert db in ["sra", "geo"]

        payload = self.esearch_params[db].copy()
        if isinstance(term, list):
            term = " OR ".join(term)
        payload += [("term", term)]
        request = requests.post(self.base_url["esearch"], data=OrderedDict(payload))
        try:
            esearch_response = request.json()
        except JSONDecodeError:
            sys.stderr.write(
                "Unable to parse esummary response json: {}{}. Will retry once.".format(
                    request.text, os.linesep
                )
            )
            retry_after = request.headers.get("Retry-After", 1)
            time.sleep(int(retry_after))
            request = requests.post(self.base_url["esearch"], data=OrderedDict(payload))
            try:
                esearch_response = request.json()
            except JSONDecodeError:
                sys.stderr.write(
                    "Unable to parse esummary response json: {}{}. Aborting.".format(
                        request.text, os.linesep
                    )
                )
                sys.exit(1)

            # retry again

        if "esummaryresult" in esearch_response:
            print("No result found")
            return
        if "error" in esearch_response:
            # API rate limite exceeded
            esearch_response = _retry_response(
                self.base_url["esearch"], payload, "esearchresult"
            )

        n_records = int(esearch_response["esearchresult"]["count"])

        results = {}
        for retstart in get_retmax(n_records):
            payload = self.esearch_params[db].copy()
            payload += self.create_esummary_params(esearch_response["esearchresult"])
            payload = OrderedDict(payload)
            payload["retstart"] = retstart
            request = requests.get(
                self.base_url["esummary"], params=OrderedDict(payload)
            )
            try:
                response = request.json()
            except JSONDecodeError:
                time.sleep(1)
                response = _retry_response(self.base_url["esummary"], payload, "result")

            if "error" in response:
                # API rate limite exceeded
                response = _retry_response(self.base_url["esummary"], payload, "result")
            if retstart == 0:
                results = response["result"]
            else:
                result = response["result"]
                for key, value in result.items():
                    if key in list(results.keys()):
                        results[key] += value
                    else:
                        results[key] = value
        return results

    def get_efetch_response(self, db, term, usehistory="y"):
        assert db in ["sra", "geo"]

        payload = self.esearch_params[db].copy()
        if isinstance(term, list):
            term = " OR ".join(term)
        payload += [("term", term)]

        request = requests.get(self.base_url["esearch"], params=OrderedDict(payload))
        esearch_response = request.json()
        if "esummaryresult" in esearch_response:
            print("No result found")
            return
        if "error" in esearch_response:
            # API rate limite exceeded
            esearch_response = _retry_response(
                self.base_url["esearch"], payload, "esearchresult"
            )

        n_records = int(esearch_response["esearchresult"]["count"])

        results = {}
        for retstart in get_retmax(n_records):
            payload = self.efetch_params.copy()
            payload += self.create_esummary_params(esearch_response["esearchresult"])
            payload = OrderedDict(payload)
            payload["retstart"] = retstart
            request = requests.get(self.base_url["efetch"], params=OrderedDict(payload))
            request_text = request.text.strip()
            try:
                request_json = request.json()
            except:
                request_json = {}  # eval(request_text)

            if "error" in request_json:
                # print("Encountered: {}".format(request_json))
                # print("Headers: {}".format(request.headers))
                # Handle API-rate limit exceeding
                try:
                    retry_after = request.headers["Retry-After"]
                except KeyError:
                    if request_json["error"] == "error forwarding request":
                        sys.stderr.write("Encountered error while making request.\n")
                        sys.exit(1)
                time.sleep(int(retry_after))
                # try again
                request = requests.get(
                    self.base_url["efetch"], params=OrderedDict(payload)
                )
                request_text = request.text.strip()
                try:
                    request_json = request.json()
                    if request_json["error"] == "error forwarding request":
                        sys.stderr.write("Encountered error while making request.\n")
                        return
                except:
                    request_json = {}  # eval(request_text)
            try:
                xml_response = xmltodict.parse(
                    request_text, dict_constructor=OrderedDict
                )

                exp_response = xml_response.get("EXPERIMENT_PACKAGE_SET", {})
                response = exp_response.get("EXPERIMENT_PACKAGE", {})
            except ExpatError:
                sys.stderr.write(
                    "Unable to parse xml: {}{}".format(request_text, os.linesep)
                )
                sys.exit(1)
            if not response:
                sys.stderr.write(
                    "Unable to parse xml response. Received: {}{}".format(
                        xml_response, os.linesep
                    )
                )
                sys.exit(1)
            if retstart == 0:
                results = response
            else:
                result = response
                for value in result:
                    results.append(value)
            time.sleep(self.sleep_time)
        return results

    def sra_metadata(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
        output_read_lengths=False,
        **kwargs
    ):
        esummary_result = self.get_esummary_response("sra", srp)
        try:
            uids = esummary_result["uids"]
        except KeyError:
            print("No results found for {}".format(srp))
            return None

        exps_xml = OrderedDict()
        runs_xml = OrderedDict()

        exps_json = OrderedDict()
        runs_json = OrderedDict()
        for uid in uids:
            exps_xml[uid] = self.format_xml(esummary_result[uid]["expxml"])
            runs_xml[uid] = self.format_xml(esummary_result[uid]["runs"])

        for uid in uids:
            exps_json[uid] = self.xml_to_json(exps_xml[uid])
            runs_json[uid] = self.xml_to_json(runs_xml[uid])

        sra_record = []
        for uid, run_json in runs_json.items():
            exp_json = exps_json[uid]
            exp_summary = exp_json["Summary"]
            exp_title = exp_summary.get("Title", pd.NA)
            exp_platform = exp_summary.get("Platform", {})
            statistics = exp_summary.get("Statistics", {})
            if isinstance(exp_platform, OrderedDict):
                exp_platform_model = exp_platform.get("@instrument_model", pd.NA)
                exp_platform_desc = exp_platform.get("#text", pd.NA)
            else:
                exp_platform_model = pd.NA
                exp_platform_desc = pd.NA

            exp_total_runs = statistics.get("@total_runs", pd.NA)
            exp_total_spots = statistics.get("@total_spots", pd.NA)
            exp_total_size = statistics.get("@total_size", pd.NA)

            # experiment_accession
            exp_ID = exp_json["Experiment"]["@acc"]
            # experiment_title
            exp_name = exp_json["Experiment"]["@name"]

            exp_organism = exp_json.get("Organism", pd.NA)
            exp_organism_name = pd.NA
            exp_taxid = pd.NA
            if isinstance(exp_organism, dict):
                exp_organism_name = exp_organism.get("@ScientificName", pd.NA)
                exp_taxid = exp_organism["@taxid"]

            exp_instrument = list(exp_json["Instrument"].values())[0]

            exp_sample = exp_json["Sample"]
            # sample_accession
            exp_sample_ID = exp_sample["@acc"]
            # sample_title
            exp_sample_name = exp_sample["@name"]

            exp_library_descriptor = exp_json["Library_descriptor"]
            # library_strategy
            exp_library_strategy = exp_library_descriptor["LIBRARY_STRATEGY"]
            if isinstance(exp_library_strategy, dict):
                exp_library_strategy = exp_library_strategy["#text"]
            # library_source
            exp_library_source = exp_library_descriptor["LIBRARY_SOURCE"]
            if isinstance(exp_library_source, dict):
                exp_library_source = exp_library_source["#text"]
            # library_selection
            exp_library_selection = exp_library_descriptor["LIBRARY_SELECTION"]
            if isinstance(exp_library_selection, dict):
                exp_library_selection = exp_library_selection["#text"]
            # library_name
            exp_library_name = exp_library_descriptor.get("LIBRARY_NAME", "")
            if isinstance(exp_library_selection, dict):
                exp_library_name = exp_library_name["#text"]
            # library_layout
            exp_library_layout = list(exp_library_descriptor["LIBRARY_LAYOUT"].keys())[
                0
            ]
            experiment_record = OrderedDict()
            experiment_record["study_accession"] = exp_json["Study"]["@acc"]
            experiment_record["study_title"] = exp_json["Study"]["@name"]
            experiment_record["experiment_accession"] = exp_ID
            experiment_record["experiment_title"] = exp_name
            experiment_record["experiment_desc"] = exp_title

            experiment_record["organism_taxid"] = exp_taxid
            experiment_record["organism_name"] = exp_organism_name

            experiment_record["library_name"] = exp_library_name
            experiment_record["library_strategy"] = exp_library_strategy
            experiment_record["library_source"] = exp_library_source
            experiment_record["library_selection"] = exp_library_selection
            experiment_record["library_layout"] = exp_library_layout
            experiment_record["sample_accession"] = exp_sample_ID
            experiment_record["sample_title"] = exp_sample_name
            experiment_record["instrument"] = exp_instrument
            experiment_record["instrument_model"] = exp_platform_model
            experiment_record["instrument_model_desc"] = exp_platform_desc
            experiment_record["total_spots"] = exp_total_spots
            experiment_record["total_size"] = exp_total_size
            if not run_json:
                # Sometimes the run_accession is not populated by NCBI:
                # df2 = self.srx_to_srr(exp_ID)
                # extra_fields = set(experiment_record.keys()).difference(df2.columns.tolist())
                # for idx, row in df2.iterrows():
                #    for field in extra_fields:
                #        experiment_record[field] = row[field]
                sra_record.append(experiment_record)
                continue
            runs = run_json["Run"]
            if not isinstance(runs, list):
                runs = [runs]
            for run_record in runs:
                run_accession = run_record["@acc"]
                run_total_spots = run_record["@total_spots"]
                run_total_bases = run_record["@total_bases"]

                experiment_record["run_accession"] = run_accession
                experiment_record["run_total_spots"] = run_total_spots
                experiment_record["run_total_bases"] = run_total_bases

                sra_record.append(experiment_record)

        # TODO: the detailed call below does redundant operations
        # the code above this can be completeley done away with
        metadata_df = (
            pd.DataFrame(sra_record).drop_duplicates().sort_values(by="run_accession")
        )
        metadata_df.columns = [x.lower().strip() for x in metadata_df.columns]
        if not detailed:
            return metadata_df

        time.sleep(self.sleep_time)
        efetch_result = self.get_efetch_response("sra", srp)
        if not isinstance(efetch_result, list):
            if efetch_result:
                efetch_result = [efetch_result]
            else:
                return None

        detailed_records = []
        for record in efetch_result:
            if "SAMPLE_ATTRIBUTES" in record["SAMPLE"]:
                sample_attributes = record["SAMPLE"]["SAMPLE_ATTRIBUTES"][
                    "SAMPLE_ATTRIBUTE"
                ]
            else:
                sample_attributes = []
            if isinstance(sample_attributes, OrderedDict):
                sample_attributes = [sample_attributes]
            exp_record = record["EXPERIMENT"]
            exp_attributes = exp_record.get("EXPERIMENT_ATTRIBUTES", {})
            run_sets = record["RUN_SET"].get("RUN", [])

            if not isinstance(run_sets, list):
                run_sets = [run_sets]

            for run_set in run_sets:
                detailed_record = OrderedDict()
                if not run_json:
                    # Add experiment accession if no run info found earlier
                    detailed_record["experiment_accession"] = exp_record["@accession"]
                # detailed_record["experiment_title"] = exp_record["TITLE"]
                for key, values in exp_attributes.items():
                    key = key.lower()
                    for value_x in values:
                        if not isinstance(value_x, dict):
                            continue
                        tag = value_x["TAG"].lower()
                        value = value_x["VALUE"]
                        detailed_record[tag] = value
                lib_record = exp_record["DESIGN"]["LIBRARY_DESCRIPTOR"]
                for key, value in lib_record.items():
                    key = key.lower()
                    if key == "library_layout":
                        value = list(value.keys())[0]
                    elif key == "library_construction_protocol":
                        continue
                    # detailed_record[key] = value

                detailed_record["run_accession"] = run_set["@accession"]
                detailed_record["run_alias"] = run_set["@alias"]
                sra_files = run_set.get("SRAFiles", {})
                sra_files = sra_files.get("SRAFile", {})
                if isinstance(sra_files, OrderedDict):
                    # detailed_record["sra_url"] = sra_files.get("@url", pd.NA)
                    if "Alternatives" in sra_files.keys():
                        alternatives = sra_files["Alternatives"]
                        if not isinstance(alternatives, list):
                            alternatives = [alternatives]
                        for alternative in alternatives:
                            org = alternative["@org"].lower()
                            for key in alternative.keys():
                                if key == "@org":
                                    continue
                                detailed_record[
                                    "{}_{}".format(org, key.replace("@", ""))
                                ] = alternative[key]

                else:
                    for sra_file in sra_files:
                        # Multiple download URLs
                        # Use the one where the download filename corresponds to the SRR
                        cluster = sra_file.get("@cluster", None).lower().strip()
                        if cluster is None:
                            continue
                        for key in sra_file.keys():
                            if key == "@cluster":
                                continue
                            if key == "Alternatives":
                                # Example: SRP184142
                                alternatives = sra_file["Alternatives"]
                                if not isinstance(alternatives, list):
                                    alternatives = [alternatives]
                                for alternative in alternatives:
                                    org = alternative["@org"].lower()
                                    for key in alternative.keys():
                                        if key == "@org":
                                            continue
                                        detailed_record[
                                            "{}_{}".format(org, key.replace("@", ""))
                                        ] = alternative[key]
                            else:
                                detailed_record[
                                    "{}_{}".format(cluster, key.replace("@", ""))
                                ] = sra_file[key]

                expt_ref = run_set["EXPERIMENT_REF"]
                detailed_record["experiment_alias"] = expt_ref.get("@refname", "")
                # detailed_record["run_total_bases"] = run_set["@total_bases"]
                # detailed_record["run_total_spots"] = run_set["@total_spots"]
                for sample_attribute in sample_attributes:
                    dict_values = list(sample_attribute.values())
                    if len(dict_values) > 1:
                        detailed_record[dict_values[0]] = dict_values[1]
                    else:
                        # TODO: Investigate why these fields have just the key
                        # but no value
                        pass
                detailed_records.append(detailed_record)
        detailed_record_df = pd.DataFrame(detailed_records).drop_duplicates()
        if (
            "run_accession" in metadata_df.keys()
            and "run_accession" in detailed_record_df.keys()
        ):
            metadata_df = metadata_df.merge(
                detailed_record_df, on="run_accession", how="outer"
            )
        else:
            metadata_df = metadata_df.merge(
                detailed_record_df, on="experiment_accession", how="outer"
            )

        metadata_df = metadata_df[metadata_df.columns.dropna()]
        metadata_df = metadata_df.drop_duplicates()
        metadata_df = metadata_df.replace(r"^\s*$", np.nan, regex=True)
        ena_cols = [
            "ena_fastq_http",
            "ena_fastq_http_1",
            "ena_fastq_http_2",
            "ena_fastq_ftp",
            "ena_fastq_ftp_1",
            "ena_fastq_ftp_2",
        ]
        empty_df = pd.DataFrame(columns=ena_cols)
        metadata_df = pd.concat((metadata_df, empty_df), axis=0)
        # metadata_df[ena_cols] = np.nan

        metadata_df = metadata_df.set_index("run_accession")
        # multithreading lookup on ENA, since a lot of time is spent waiting
        # for its reply
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # load our function calls into a list of futures
            futures = [
                executor.submit(self.fetch_ena_fastq, srp)
                for srp in metadata_df.study_accession.unique()
            ]
            # now proceed synchronously
            for future in concurrent.futures.as_completed(futures):
                ena_results = future.result()
                if ena_results.shape[0]:
                    ena_results = ena_results.set_index("run_accession")
                    metadata_df.update(ena_results)
        metadata_df = metadata_df.reset_index()
        metadata_df = metadata_df.fillna(pd.NA)
        metadata_df.columns = [x.lower().strip() for x in metadata_df.columns]
        return metadata_df.sort_values(by="run_accession")

    def fetch_gds_results(self, gse, **kwargs):
        result = self.get_esummary_response("geo", gse)

        try:
            uids = result["uids"]
        except KeyError:
            print("No results found for {} | Obtained result: {}".format(gse, result))
            return None
        gse_records = []
        for uid in uids:
            record = result[uid]
            del record["uid"]
            if record["extrelations"]:
                extrelations = record["extrelations"]
                for extrelation in extrelations:
                    keys = list(extrelation.keys())
                    values = list(extrelation.values())
                    assert sorted(keys) == sorted(
                        ["relationtype", "targetobject", "targetftplink"]
                    )
                    assert len(values) == 3
                    record[extrelation["relationtype"]] = extrelation["targetobject"]
                del record["extrelations"]
                gse_records.append(record)
        if not len(gse_records):
            print("No results found for {}".format(gse))
            return None
        return pd.DataFrame(gse_records)

    def gse_to_gsm(self, gse, **kwargs):
        gse_df = self.fetch_gds_results(gse, **kwargs)
        gse_df = gse_df.rename(
            columns={
                "accession": "experiment_alias",
                "SRA": "experiment_accession",
                "title": "experiment_title",
                "summary": "sample_attribute",
            }
        )
        # TODO: Fix for multiple GSEs?
        gse_df["study_alias"] = ""
        for index, row in gse_df.iterrows():
            if row.entrytype == "GSE":
                study_alias = row["experiment_accession"]
            # If GSM is ecnountered, apply it the
            # previously encountered GSE
            elif row.entrytype == "GSM":
                gse_df.loc[index, "study_alias"] = study_alias
        gse_df = gse_df[gse_df.entrytype == "GSM"]
        if kwargs and kwargs["detailed"] == True:
            return gse_df
        return gse_df[
            ["study_alias", "experiment_alias", "experiment_accession"]
        ].drop_duplicates()

    def gse_to_srp(self, gse, **kwargs):
        gse_df = self.fetch_gds_results(gse, **kwargs)
        gse_df = gse_df[gse_df.entrytype == "GSE"]
        gse_df = gse_df.rename(
            columns={"accession": "study_alias", "SRA": "study_accession"}
        )
        return gse_df[["study_alias", "study_accession"]].drop_duplicates()

    def gsm_to_srp(self, gsm, **kwargs):
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSE"]
        gsm_df = gsm_df.rename(
            columns={"accession": "experiment_alias", "SRA": "study_accession"}
        )
        return gsm_df[["experiment_alias", "study_accession"]].drop_duplicates()

    def gsm_to_srr(self, gsm, **kwargs):
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df.rename(
            columns={
                "accession": "experiment_alias",
                "SRA": "experiment_accession",
                "title": "experiment_title",
                "summary": "sample_attribute",
            }
        )
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"]
        srr_df = self.srx_to_srr(gsm_df.experiment_accession.tolist())
        gsm_df = gsm_df.merge(srr_df, on="experiment_accession")
        return gsm_df[["experiment_alias", "run_accession"]]

    def gsm_to_srs(self, gsm, **kwargs):
        """Get SRS for a GSM"""
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        srx = gsm_df.experiment_accession.tolist()
        time.sleep(self.sleep_time)
        srs_df = self.srx_to_srs(srx)
        gsm_df = srs_df.merge(gsm_df, on="experiment_accession")[
            ["experiment_alias", "sample_accession"]
        ]
        return gsm_df.drop_duplicates()

    def gsm_to_srx(self, gsm, **kwargs):
        """Get SRX for a GSM"""
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        return gsm_df[["experiment_alias", "experiment_accession"]].drop_duplicates()

    def gsm_to_gse(self, gsm, **kwargs):
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSE"]
        gsm_df = gsm_df.rename(
            columns={"accession": "study_alias", "SRA": "study_accession"}
        )
        return gsm_df[["study_alias", "study_accession"]]

    def srp_to_gse(self, srp, **kwargs):
        """Get GSE for a SRP"""
        srp_df = self.fetch_gds_results(srp, **kwargs)
        srp_df = srp_df[srp_df.entrytype == "GSE"]
        srp_df = srp_df.rename(
            columns={"accession": "study_alias", "SRA": "study_accession"}
        )
        return srp_df[["study_accession", "study_alias"]].drop_duplicates()

    def srp_to_srr(self, srp, **kwargs):
        """Get SRR for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        return _order_first(srp_df, ["study_accession", "run_accession"])

    def srp_to_srs(self, srp, **kwargs):
        """Get SRS for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        return _order_first(srp_df, ["study_accession", "sample_accession"])

    def srp_to_srx(self, srp, **kwargs):
        """Get SRX for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        srp_df["study_accesssion"] = srp
        return _order_first(srp_df, ["study_accession", "experiment_accession"])

    def srr_to_gsm(self, srr, **kwargs):
        """Get GSM for a SRR"""
        srr_df = self.srr_to_srp(srr, detailed=True)
        # remove NAs
        srp = [x for x in srr_df.study_accession.tolist() if not x is pd.NA]
        gse_df = self.fetch_gds_results(srp, **kwargs)
        gse_df = gse_df[gse_df.entrytype == "GSE"].rename(
            columns={"SRA": "project_accession", "accession": "project_alias"}
        )
        gsm_df = self.gse_to_gsm(gse_df.project_alias.tolist(), detailed=True)
        srr_cols = list(
            set(srr_df.columns.tolist()).difference(gsm_df.columns.tolist())
        ) + ["experiment_accession"]
        joined_df = gsm_df.merge(srr_df[srr_cols], on="experiment_accession")
        return _order_first(joined_df, ["run_accession", "experiment_alias"])

    def srr_to_srp(self, srr, **kwargs):
        """Get SRP for a SRR"""
        srr_df = self.sra_metadata(srr, **kwargs)
        if kwargs and kwargs["detailed"] == True:
            return srr_df
        return _order_first(srr_df, ["run_accession", "study_accession"])

    def srr_to_srs(self, srr, **kwargs):
        """Get SRS for a SRR"""
        srr_df = self.sra_metadata(srr, **kwargs)
        return _order_first(srr_df, ["run_accession", "sample_accession"])

    def srr_to_srx(self, srr, **kwargs):
        """Get SRX for a SRR"""
        srr_df = self.sra_metadata(srr)
        return _order_first(srr_df, ["run_accession", "experiment_accession"])

    def srs_to_gsm(self, srs, **kwargs):
        """Get GSM for a SRS"""
        srx_df = self.srs_to_srx(srs)
        time.sleep(self.sleep_time)
        gsm_df = self.srx_to_gsm(srx_df.experiment_accession.tolist(), **kwargs)
        srs_df = srx_df.merge(gsm_df, on="experiment_accession")
        return _order_first(srs_df, ["sample_accession", "experiment_alias"])

    def srx_to_gsm(self, srx, **kwargs):
        gsm_df = self.fetch_gds_results(srx, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        return gsm_df[["experiment_accession", "experiment_alias"]]

    def srs_to_srx(self, srs, **kwargs):
        """Get SRX for a SRS"""
        srs_df = self.sra_metadata(srs, **kwargs)
        return _order_first(srs_df, ["sample_accession", "experiment_accession"])

    def srx_to_srp(self, srx, **kwargs):
        """Get SRP for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "study_accession"])

    def srx_to_srr(self, srx, **kwargs):
        """Get SRR for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "run_accession"])

    def srx_to_srs(self, srx, **kwargs):
        """Get SRS for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "sample_accession"])

    def search(self, *args, **kwargs):
        raise NotImplementedError("Search not yet implemented for Web")

    def close(self):
        # Dummy method to mimick SRAdb() object
        # TODO: fix this
        pass



"""Tests for search.py
"""

import json

import pandas as pd
import pytest

from pysradb.search import *

# =========================== pytest fixtures ============================
# pytest fixtures containing simulated input/outputs for various functions
# used in the search module

# Search inputs are packaged in lists in the following formats so that they
# can be passed to test functions easily:
# Format for Sra/Ena: [verbosity, return_max, query, accession, organism,
# layout, mbases, publication_date, platform, selection, source, strategy,
# title, suppress_validations]
# Format for Geo: [verbosity, return_max, query, accession, organism,
# layout, mbases, publication_date, platform, selection, source, strategy,
# title, geo_query, geo_dataset_type, geo_entry_type, suppress_validations]


@pytest.fixture(scope="module")
def valid_search_inputs_1():
    """Basic search input tests for Sra and Ena (single inputs)"""
    return [
        [
            0,
            20,
            ["covid-19"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # verbosity
        [
            3,
            20,
            ["covid-19"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # query
        [
            2,
            20,
            None,
            "SRS6898940",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # accession
        [
            2,
            20,
            None,
            None,
            "Escherichia coli",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # organism
        [
            2,
            20,
            None,
            None,
            None,
            "PAIRED",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # layout
        [
            2,
            20,
            None,
            None,
            None,
            None,
            5,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # mbases
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            "01-01-2019:31-12-2019",
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # pdat
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            "ion torrent",
            None,
            None,
            None,
            None,
            False,
        ],  # platform
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "random",
            None,
            None,
            None,
            False,
        ],  # selection
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "genomic",
            None,
            None,
            False,
        ],  # source
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "wgs",
            None,
            False,
        ],  # strategy
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "Homo sapiens; RNA-Seq",
            False,
        ],  # title
    ]


@pytest.fixture(scope="module")
def valid_search_inputs_2():
    """More complex input tests for Sra and Ena"""
    return [
        [
            0,
            20,
            None,
            None,
            None,
            "Triple",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            True,
        ],  # suppress_validation
        [
            2,
            20,
            ["Escherichia", "coli"],
            "SRS6898222",
            "Escherichia coli",
            "paired",
            6,
            "01-01-1999:31-12-2019",
            "ILLUMINA",
            "dnase",
            "metatranscriptomic",
            "mbd seq",
            None,
            False,
        ],
        [
            1,
            2,
            [],
            None,
            None,
            "single",
            "5",
            "31-12-2019",
            "Nanopore",
            "MBD2 protein methyl-CpG binding domain",
            "genomic single cell",
            "amplicon selection",
            None,
            False,
        ],
        [
            0,
            200000,
            None,
            None,
            None,
            "Paired",
            None,
            None,
            "complete genomics",
            "Inverse rRNA",
            "TRANSCRIPTOMIC",
            "hi-c",
            None,
            False,
        ],
        [
            3,
            1,
            None,
            None,
            None,
            "SINGLE",
            None,
            None,
            "ls454",
            "oligo-dt",
            "transcriptomic single-cell",
            "miRNA",
            None,
            False,
        ],
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            "smrt",
            "cDNA_oligo_dT",
            "metagenomic",
            "MBD",
            None,
            False,
        ],
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            "Pacbio",
            "pcr",
            "others",
            "EST",
            None,
            False,
        ],
    ]


@pytest.fixture(scope="module")
def valid_search_inputs_geo():
    """Basic search input tests for Geo"""
    return [
        [
            2,
            20,
            "query",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "GEO query",
            None,
            None,
            False,
        ],
        [
            2,
            20,
            ["query"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "GEO query",
            None,
            None,
            False,
        ],
        [
            2,
            20,
            None,
            "SRS6898940",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "GEO query",
            None,
            None,
            False,
        ],
        [
            2,
            20,
            None,
            None,
            "Escherichia coli",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # organism
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            "01-01-2019:31-12-2019",
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # pdat
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "GEO dataset type",
            None,
            False,
        ],  # Geo dataset type
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "GEO entry type",
            False,
        ],
    ]


@pytest.fixture(scope="module")
def empty_search_inputs():
    return [
        [],
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],
        [2, 20, [], [], [], [], [], [], [], [], [], [], []],
    ]


@pytest.fixture(scope="module")
def empty_search_inputs_geo():
    return [
        [],
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],
        [2, 20, [], [], [], [], [], [], [], [], [], [], []],
    ]


@pytest.fixture(scope="module")
def invalid_search_inputs():
    """Invalid search input tests for QuerySearch"""
    return [
        [
            4,
            20,
            ["covid-19"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # verbosity
        [
            3,
            0,
            ["covid-19"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # return_max
        [
            2,
            20,
            None,
            None,
            None,
            "X",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # layout
        [
            2,
            20,
            None,
            None,
            None,
            None,
            "Charmander",
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # mbases
        [
            2,
            20,
            None,
            None,
            None,
            None,
            -1,
            None,
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # mbases
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            "31-31-2019",
            None,
            None,
            None,
            None,
            None,
            False,
        ],  # pdat
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            "pacbio nanopore",
            None,
            None,
            None,
            None,
            False,
        ],  # platform
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            "no such platform",
            None,
            None,
            None,
            None,
            False,
        ],  # platform
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "polyA hybrid",
            None,
            None,
            None,
            False,
        ],  # selection
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "no such selection",
            None,
            None,
            None,
            False,
        ],  # selection
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "genomic transcriptomic",
            None,
            None,
            False,
        ],  # source
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "metagenomic viral rna ",
            None,
            None,
            False,
        ],  # source
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "wgs wga",
            None,
            False,
        ],  # strategy
        [
            2,
            20,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "Bulbasaur",
            None,
            False,
        ],  # strategy
    ]


@pytest.fixture(scope="module")
def sra_response_xml_1():
    return "./tests/data/test_search/sra_test.xml"


@pytest.fixture(scope="module")
def sra_formatted_responses_1():
    return [
        pd.read_csv(
            "./tests/data/test_search/sra_test_verbosity_0.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_verbosity_1.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_verbosity_2.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_verbosity_3.csv",
            dtype=object,
            keep_default_na=False,
        ),
    ]


@pytest.fixture(scope="module")
def sra_response_xml_2():
    return "./tests/data/test_search/sra_test_ERS3331676.xml"


@pytest.fixture(scope="module")
def sra_formatted_responses_2():
    return [
        pd.read_csv(
            "./tests/data/test_search/sra_test_2_verbosity_0.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_2_verbosity_1.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_2_verbosity_2.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/sra_test_2_verbosity_3.csv",
            dtype=object,
            keep_default_na=False,
        ),
    ]


@pytest.fixture(scope="module")
def sra_uids():
    with open("./tests/data/test_search/sra_uids.txt", "r") as f:
        uids = f.read().splitlines()
    return uids


@pytest.fixture(scope="module")
def ena_responses_json():
    data = []
    for i in range(4):
        with open(f"./tests/data/test_search/ena_test_verbosity_{i}.json") as f:
            data.append(json.load(f))
    return data


@pytest.fixture(scope="module")
def ena_formatted_responses():
    return [
        pd.read_csv(
            "./tests/data/test_search/ena_test_verbosity_0.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/ena_test_verbosity_1.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/ena_test_verbosity_2.csv",
            dtype=object,
            keep_default_na=False,
        ),
        pd.read_csv(
            "./tests/data/test_search/ena_test_verbosity_3.csv",
            dtype=object,
            keep_default_na=False,
        ),
    ]


# ============================= General Tests =============================


def missing_query_test(empty_search_inputs):
    for empty_search_input in empty_search_inputs:
        with pytest.raises(MissingQueryException):
            QuerySearch(*empty_search_input)


def test_invalid_search_query(invalid_search_inputs):
    error_messages = [
        "Incorrect verbosity format",
        "Incorrect return_max format",
        "Incorrect layout field format",
        "Incorrect mbases format",
        "Incorrect mbases format",
        "Incorrect publication date format",
        "Multiple potential matches have been identified for pacbio nanopore",
        "Incorrect platform",
        "Multiple potential matches have been identified for polyA hybrid",
        "Incorrect selection",
        "Incorrect source",
        "Multiple potential matches have been identified for metagenomic viral rna",
        "Multiple potential matches have been identified for wgs wga",
        "Incorrect strategy",
    ]
    for i in range(len(invalid_search_inputs)):
        with pytest.raises(IncorrectFieldException) as e:
            QuerySearch(*invalid_search_inputs[i])
        assert error_messages[i] in str(e.value)


# ======================== SraSearch component tests ======================


def test_sra_search_1():
    instance = SraSearch(
        3, 1000, query="ribosome profiling", publication_date="01-10-2012:01-01-2013"
    )
    instance.search()
    found_accessions = set(instance.get_df()["experiment_accession"])
    with open("./tests/data/test_search/sra_search_test1.txt", "r") as f:
        expected_accessions = f.read().splitlines()
    assert found_accessions == set(expected_accessions)


def test_sra_uids(sra_uids):
    instance = SraSearch(
        3, 1000, query="ribosome profiling", publication_date="01-10-2012:01-01-2013"
    )
    instance.search()
    assert instance.get_uids() == sra_uids


def test_valid_search_query_1_sra(valid_search_inputs_1):
    expected_query = [
        "covid-19",
        "covid-19",
        "SRS6898940[Accession]",
        "Escherichia coli[Organism]",
        "PAIRED[Layout]",
        "5[Mbases]",
        "2019/01/01:2019/12/31[PDAT]",
        "ION_TORRENT[Platform]",
        "RANDOM[Selection]",
        "GENOMIC[Source]",
        "WGS[Strategy]",
        "Homo sapiens; RNA-Seq[Title]",
    ]
    for i in range(len(valid_search_inputs_1)):
        assert (
            SraSearch(*valid_search_inputs_1[i])._format_query_string()
            == expected_query[i]
        )


def test_valid_search_query_2_sra(valid_search_inputs_2):
    expected_query = [
        "Triple[Layout]",
        "Escherichia coli AND SRS6898222[Accession] AND Escherichia coli[Organism] AND paired[Layout] AND "
        "6[Mbases] AND 1999/01/01:2019/12/31[PDAT] AND ILLUMINA[Platform] AND DNase[Selection] AND "
        "METATRANSCRIPTOMIC[Source] AND MBD-Seq[Strategy]",
        "single[Layout] AND 5[Mbases] AND 2019/12/31[PDAT] AND OXFORD_NANOPORE[Platform] AND "
        "MBD2 protein methyl-CpG binding domain[Selection] AND GENOMIC SINGLE CELL[Source] AND AMPLICON[Strategy]",
        "Paired[Layout] AND COMPLETE_GENOMICS[Platform] AND Inverse rRNA[Selection] AND TRANSCRIPTOMIC[Source] AND "
        "Hi-C[Strategy]",
        "SINGLE[Layout] AND LS454[Platform] AND Oligo-dT[Selection] AND TRANSCRIPTOMIC SINGLE CELL[Source] AND "
        "miRNA-Seq[Strategy]",
        "PACBIO_SMRT[Platform] AND cDNA_oligo_dT.*[Selection] AND METAGENOMIC[Source] AND MBD-Seq[Strategy]",
        "PACBIO_SMRT[Platform] AND PCR[Selection] AND OTHER[Source] AND EST[Strategy]",
    ]
    for i in range(len(valid_search_inputs_2)):
        assert (
            SraSearch(*valid_search_inputs_2[i])._format_query_string()
            == expected_query[i]
        )


def test_sra_search_format_request():
    assert SraSearch(
        0,
        1000,
        query="covid-19",
    )._format_request() == {
        "db": "sra",
        "term": "covid-19",
        "retmode": "json",
        "retmax": 1000,
    }


def test_sra_search_format_result_1(sra_response_xml_1, sra_formatted_responses_1):
    for i in range(2, 4):
        query = SraSearch(
            i,
            1000,
            query="ribosome profiling",
            platform="illumina",
            organism="Caenorhabditis elegans",
        )
        query._format_response(sra_response_xml_1)
        query._format_result()
        col0 = [
            c
            for c in query.get_df().columns
            if ("run" not in c.lower() and "sample" not in c.lower())
        ]
        col1 = [
            c
            for c in sra_formatted_responses_1[i].columns
            if ("run" not in c.lower() and "sample" not in c.lower())
        ]
        expected_df = (
            sra_formatted_responses_1[i][col1].fillna("N/A").replace("<NA>", "N/A")
        )
        actual_df = query.get_df()[col0].fillna("N/A").replace("<NA>", "N/A")
        pd.testing.assert_frame_equal(expected_df, actual_df, check_dtype=False)


def test_sra_search_format_result_2(sra_response_xml_2, sra_formatted_responses_2):
    for i in range(4):
        query = SraSearch(i, 1000, accession="ERS3331676")

        query._format_response(sra_response_xml_2)
        query._format_result()
        col0 = [
            c
            for c in query.get_df().columns
            if ("run" not in c.lower() and "sample" not in c.lower())
        ]
        col1 = [
            c
            for c in sra_formatted_responses_2[i].columns
            if ("run" not in c.lower() and "sample" not in c.lower())
        ]
        expected_df = (
            sra_formatted_responses_2[i][col1].fillna("N/A").replace("<NA>", "N/A")
        )
        actual_df = query.get_df()[col0].fillna("N/A").replace("<NA>", "N/A")
        pd.testing.assert_frame_equal(expected_df, actual_df, check_dtype=False)


# ====================== EnaSearch component tests ========================


def test_ena_search_1():
    instance = EnaSearch(
        0, 1000, platform="pacbio", publication_date="01-10-2012:01-01-2013"
    )
    instance.search()
    df = instance.get_df()["run_accession"].to_list()
    with open("./tests/data/test_search/ena_search_test1.txt", "r") as f:
        expected_accessions = f.read().splitlines()
    for accession in df:
        assert accession in expected_accessions


def test_ena_search_2(capsys):
    EnaSearch(0, 1000, query="hehehuhuhaha").search()
    out, err = capsys.readouterr()
    assert "No results found for the following search query:" in out
    assert err == ""


def test_ena_search_3(capsys):
    with pytest.raises(SystemExit) as e:
        EnaSearch(0, 1000, selection='"Pikachu', suppress_validation=True).search()
    assert "HTTPError: This is likely caused by an invalid search query:" in str(
        e.value
    )


def test_valid_search_query_1_ena(valid_search_inputs_1):
    expected_query = [
        '(experiment_title="*covid-19*" OR study_accession="COVID-19" OR secondary_study_accession="COVID-19" OR'
        ' sample_accession="COVID-19" OR secondary_sample_accession="COVID-19" OR experiment_accession="COVID-19" OR'
        ' submission_accession="COVID-19" OR run_accession="COVID-19")',
        '(experiment_title="*covid-19*" OR study_accession="COVID-19" OR secondary_study_accession="COVID-19" OR'
        ' sample_accession="COVID-19" OR secondary_sample_accession="COVID-19" OR experiment_accession="COVID-19" OR'
        ' submission_accession="COVID-19" OR run_accession="COVID-19")',
        '(study_accession="SRS6898940" OR secondary_study_accession="SRS6898940" OR sample_accession="SRS6898940" OR'
        ' secondary_sample_accession="SRS6898940" OR experiment_accession="SRS6898940" OR'
        ' submission_accession="SRS6898940" OR run_accession="SRS6898940")',
        "tax_eq(562)",
        'library_layout="PAIRED"',
        "base_count>=4500000 AND base_count<5500000",
        "first_created>=2019-01-01 AND first_created<=2019-12-31",
        'instrument_platform="ION_TORRENT"',
        'library_selection="RANDOM"',
        'library_source="GENOMIC"',
        'library_strategy="WGS"',
        'experiment_title="*Homo sapiens; RNA-Seq*"',
    ]
    for i in range(len(valid_search_inputs_1)):
        assert (
            EnaSearch(*valid_search_inputs_1[i])._format_query_string()
            == expected_query[i]
        )


def test_valid_search_query_2_ena(valid_search_inputs_2):
    expected_query = [
        'library_layout="TRIPLE"',
        '(experiment_title="*Escherichia coli*") AND (study_accession="SRS6898222" OR '
        'secondary_study_accession="SRS6898222" OR sample_accession="SRS6898222" OR '
        'secondary_sample_accession="SRS6898222" OR experiment_accession="SRS6898222" OR '
        'submission_accession="SRS6898222" OR run_accession="SRS6898222") AND tax_eq(562) AND library_layout="PAIRED" '
        "AND base_count>=5500000 AND base_count<6500000 AND first_created>=1999-01-01 AND "
        'first_created<=2019-12-31 AND instrument_platform="ILLUMINA" AND library_selection="DNase" AND '
        'library_source="METATRANSCRIPTOMIC" AND library_strategy="MBD-Seq"',
        'library_layout="SINGLE" AND base_count>=4500000 AND base_count<5500000 AND first_created=2019-12-31 AND '
        'instrument_platform="OXFORD_NANOPORE" AND library_selection="MBD2 protein methyl-CpG binding domain" AND '
        'library_source="GENOMIC SINGLE CELL" AND library_strategy="AMPLICON"',
        'library_layout="PAIRED" AND instrument_platform="COMPLETE_GENOMICS" AND library_selection="Inverse rRNA" AND '
        'library_source="TRANSCRIPTOMIC" AND library_strategy="Hi-C"',
        'library_layout="SINGLE" AND instrument_platform="LS454" AND library_selection="Oligo-dT" AND '
        'library_source="TRANSCRIPTOMIC SINGLE CELL" AND library_strategy="miRNA-Seq"',
        'instrument_platform="PACBIO_SMRT" AND library_selection="cDNA_oligo_dT.*" AND '
        'library_source="METAGENOMIC" AND library_strategy="MBD-Seq"',
        'instrument_platform="PACBIO_SMRT" AND library_selection="PCR" AND library_source="OTHER" AND '
        'library_strategy="EST"',
    ]

    for i in range(len(valid_search_inputs_2)):
        assert (
            EnaSearch(*valid_search_inputs_2[i])._format_query_string()
            == expected_query[i]
        )


def test_ena_search_format_request():
    query_string = (
        '(experiment_title="*covid-19*" OR study_accession="COVID-19" OR '
        'secondary_study_accession="COVID-19" OR sample_accession="COVID-19" OR '
        'secondary_sample_accession="COVID-19" OR experiment_accession="COVID-19" OR '
        'submission_accession="COVID-19" OR run_accession="COVID-19")'
    )

    assert EnaSearch(2, 20, "covid-19")._format_request() == {
        "query": query_string,
        "result": "read_run",
        "format": "json",
        "limit": 20,
        "fields": "study_accession,experiment_accession,experiment_title,description,tax_id,scientific_name,"
        "library_strategy,library_source,library_selection,sample_accession,sample_title,"
        "instrument_model,run_accession,read_count,base_count,first_public,library_layout,instrument_platform",
    }


def test_ena_search_format_result(ena_responses_json, ena_formatted_responses):
    for i in range(4):
        query = EnaSearch(
            i,
            1000,
            query="ribosome profiling",
            platform="illumina",
            organism="Caenorhabditis elegans",
        )
        query._format_result(ena_responses_json[i])
        expected_df = ena_formatted_responses[i].fillna("N/A").replace("<NA>", "N/A")
        actual_df = query.get_df().fillna("N/A").replace("<NA>", "N/A")
        pd.testing.assert_frame_equal(expected_df, actual_df, check_dtype=False)


# ====================== GeoSearch component tests ========================


def missing_query_test_geo(empty_search_inputs_geo):
    for empty_search_input in empty_search_inputs_geo:
        with pytest.raises(MissingQueryException):
            QuerySearch(*empty_search_input)


def test_geo_search_1():
    instance = GeoSearch(3, 1000, geo_query="ferret")
    instance.search()
    df = instance.get_df()["experiment_accession"].to_list()
    assert len(df) > 10
    # with open("./tests/data/test_search/geo_search_test1.txt", "r") as f:
    #    expected_accessions = f.read().splitlines()

    # assert len(set(expected_accessions).intersection(df)) > 1


def test_valid_search_query_geo(valid_search_inputs_geo):
    expected_sra_query = [
        "query AND sra gds[Filter]",
        "",
        "query AND sra gds[Filter]",
        "sra gds[Filter] AND SRS6898940[Accession]",
        "sra gds[Filter] AND Escherichia coli[Organism]",
        "sra gds[Filter] AND 2019/01/01:2019/12/31[PDAT]",
        "",
        "",
    ]
    expected_geo_query = [
        "",
        "GEO query AND gds sra[Filter]",
        "GEO query AND gds sra[Filter]",
        "GEO query AND gds sra[Filter]",
        "gds sra[Filter] AND Escherichia coli[Organism]",
        "gds sra[Filter] AND 2019/01/01:2019/12/31[PDAT]",
        "gds sra[Filter] AND GEO dataset type[DataSet Type]",
        "gds sra[Filter] AND GEO entry type[Entry Type]",
    ]

    for i in range(len(valid_search_inputs_geo)):
        instance = GeoSearch(*valid_search_inputs_geo[i])
        assert instance._format_query_string() == expected_sra_query[i]
        assert instance._format_geo_query_string() == expected_geo_query[i]


def test_geo_search_format_request():
    assert GeoSearch(
        0,
        1000,
        query="covid-19",
    )._format_request() == {
        "db": "sra",
        "term": "covid-19 AND sra gds[Filter]",
        "retmode": "json",
        "retmax": 1000,
    }


def test_geo_info():
    assert type(GeoSearch.info()) == str and GeoSearch.info().startswith(
        "General Information:"
    )



"""Methods to interact with SRA"""

# This is now defunct and will be removed in a future release.

import gzip
import os
import re
import sys

from .basedb import BASEdb
from .utils import _get_url
from .utils import copyfileobj
from .utils import get_gzip_uncompressed_size

PY3 = True
if sys.version_info[0] < 3:
    PY3 = False

GEOmetadb_URL = "http://starbuck1.s3.amazonaws.com/sradb/GEOmetadb.sqlite.gz"


def download_geodb_file(download_dir=os.getcwd(), overwrite=True):
    """Download GEOmetadb.sqlite file.

    Parameters
    ----------
    download_dir: string
                  Directory to download SRAmetadb.sqlite
    overwrite: bool
               overwrite existing file(s).
               Set to True by default.

    """
    download_location = os.path.join(download_dir, "GEOmetadb.sqlite.gz")
    download_location_unzip = download_location.rstrip(".gz")

    if os.path.isfile(download_location) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".forma(
                download_location
            )
        )
    if os.path.isfile(download_location_unzip) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".format(
                download_location_unzip
            )
        )

    try:
        _get_url(GEOmetadb_URL, download_location)
    except Exception as e:
        raise RuntimeError(
            "Could not use {}.\nException: {}.\n".format(GEOmetadb_URL, e)
        )
    print("Extracting {} ...".format(download_location))
    filesize = get_gzip_uncompressed_size(download_location)
    with gzip.open(download_location, "rb") as fh_in:
        with open(download_location_unzip, "wb") as fh_out:
            copyfileobj(
                fh_in,
                fh_out,
                filesize=filesize,
                desc="Extracting {}".format("GEOmetadb.sqlite.gz"),
            )
    print("Done!")
    db = GEOdb(download_location_unzip)
    metadata = db.query("SELECT * FROM metaInfo")
    db.close()
    print("Metadata associated with {}:".format(download_location_unzip))
    print(metadata)


class GEOdb(BASEdb):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        super(GEOdb, self).__init__(sqlite_file)
        self._db_type = "GEO"
        self.valid_in_type = ["GSE", "GPL", "GSM", "GDS"]

    def gse_metadata(self, gse):
        """Get metadata for GSE ID.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gse WHERE gse='{}';".format(gse))

    def gsm_metadata(self, gsm):
        """Get metadata for GSM ID.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gsm WHERE gsm='{}';".format(gsm))

    def geo_convert(self, from_acc):
        """Convert one GEO accession to other.

        Parameters
        ----------
        from_acc: string
                  GPL/GSE/GSM accession ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query(
            "SELECT * FROM geoConvert WHERE from_acc='{}';".format(from_acc)
        )

    def gse_to_gsm(self, gse):
        """Fetch GSMs for a GSE.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query("SELECT * FROM gse_gsm WHERE gse='{}'".format(gse))

    def gsm_to_gse(self, gsm):
        """Fetch GSE for a GSM.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        mapping_df = self.query("SELECT * FROM gse_gsm WHERE gsm='{}'".format(gsm))
        return mapping_df.loc[:, ["gsm", "gse"]]

    def guess_srp_from_gse(self, gse):
        """Convert GSE to SRP id.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        srp: string
             SRP ID
        """
        results = self.query('SELECT * FROM gse WHERE gse = "' + gse + '"')
        if results.shape[0] == 1:
            supp_file = results["supplementary_file"][0]
            if supp_file:
                splitted = supp_file.split(";")
                if len(splitted):
                    match = re.findall("SRP.*", splitted[-1])
                    if len(match):
                        srp = match[0].split("/")[-1]
                        return srp
        return None
